import os

import pytest
from traitlets import TraitError

from ..core import BaseWWTWidget, DEFAULT_SURVEYS_URL

WWT_HTML_FILE = os.path.join(os.path.dirname(__file__), '..', 'nbextension', 'static', 'wwt.html')

with open(WWT_HTML_FILE) as f:
    WWT_HTML = f.read()


def test_settings_in_html():

    # Test to make sure that the initial settings listed in wwt.html match
    # the defaults for the traits.

    widget = BaseWWTWidget()
    for trait_name in sorted(widget.trait_names()):
        wwt_name = widget.trait_metadata(trait_name, 'wwt')
        if wwt_name is not None:
            value = getattr(widget, trait_name)
            if hasattr(value, 'unit'):
                cmd = 'wwt.settings.set_{0}({1!r})'.format(wwt_name, value.value)
            elif isinstance(value, bool):
                cmd = 'wwt.settings.set_{0}({1})'.format(wwt_name, str(value).lower())
            else:
                cmd = 'wwt.settings.set_{0}({1!r})'.format(wwt_name, str(value))
            assert cmd in WWT_HTML

    # Check traits that are connected to methods

    cmd = 'wwt.setForegroundImageByName({0!r})'.format(str(widget.foreground))
    assert cmd in WWT_HTML

    cmd = 'wwt.setBackgroundImageByName({0!r})'.format(str(widget.background))
    assert cmd in WWT_HTML

    cmd = 'wwt.setForegroundOpacity({0!r})'.format(widget.foreground_opacity * 100)
    assert cmd in WWT_HTML

    cmd = 'wwt.loadImageCollection({0!r})'.format(DEFAULT_SURVEYS_URL)
    assert cmd in WWT_HTML


def test_color_validation():

    widget = BaseWWTWidget()

    widget.constellation_figure_color = 'red'
    assert widget.constellation_figure_color == '#ff0000'

    widget.constellation_figure_color = '#ff0000'
    assert widget.constellation_figure_color == '#ff0000'

    widget.constellation_figure_color = (1, 0, 0)
    assert widget.constellation_figure_color == '#ff0000'

    with pytest.raises(TraitError) as exc:
        widget.constellation_figure_color = (1, 0, 0, 2)
    assert exc.value.args[0] == 'color must be a string or a tuple of 3 floats'



import sys
import uuid

if sys.version_info[0] == 2:  # noqa
    from io import BytesIO as StringIO
else:
    from io import StringIO

import warnings
from base64 import b64encode

from astropy import units as u

from traitlets import HasTraits, validate, observe
from .traits import Any, Unicode, Float, Color

__all__ = ['LayerManager', 'TableLayer']

VALID_LON_UNITS = {u.deg: 'degrees',
                   u.hour: 'hours',
                   u.hourangle: 'hours'}


# NOTE: for cartesian coordinates, we can also allow custom units
VALID_ALT_UNITS = {u.m: 'meters',
                   u.imperial.foot: 'feet',
                   u.imperial.inch: 'inches',
                   u.imperial.mile: 'miles',
                   u.km: 'kilometers',
                   u.au: 'astronomicalUnits',
                   u.lyr: 'lightYears',
                   u.pc: 'parsecs',
                   u.Mpc: 'megaParsecs'}

VALID_ALT_TYPES = ['depth', 'altitude', 'distance', 'seaLevel', 'terrain']


def guess_lon_lat_columns(colnames):
    """
    Given column names in a table, return the columns to use for lon/lat, or
    None/None if no high confidence possibilities.
    """

    # Do all the checks in lowercase
    colnames_lower = [colname.lower() for colname in colnames]

    for lon, lat in [('ra', 'dec'), ('lon', 'lat'), ('lng', 'lat')]:

        # Check first for exact matches
        if colnames_lower.count(lon) == 1 and colnames_lower.count(lat) == 1:
            return lon, lat

        # Next check for columns that start with specified names

        lon_match = [colname.startswith(lon) for colname in colnames_lower]
        lat_match = [colname.startswith(lat) for colname in colnames_lower]

        if sum(lon_match) == 1 and sum(lat_match) == 1:
            return colnames[lon_match.index(True)], colnames[lat_match.index(True)]

        # We don't check for cases where lon/lat are inside the name but not at
        # the start since that might be e.g. for proper motions (pm_ra) or
        # errors (dlat).

    return None, None


def pick_unit_if_available(unit, valid_units):
    # Check for equality rather than just identity
    for valid_unit in valid_units:
        if unit == valid_unit:
            return valid_unit
    return unit


class LayerManager(object):
    """
    A simple container for layers.
    """

    def __init__(self, parent=None):
        self._layers = []
        self._parent = parent

    def add_data_layer(self, table=None, frame='Sky', **kwargs):
        """
        Add a data layer to the current view

        Parameters
        ----------
        """
        if table is not None:
            layer = TableLayer(self._parent, table=table, frame=frame, **kwargs)
        else:
            # NOTE: in future we may allow different arguments such as e.g.
            # orbit=, hence why we haven't made this a positional argument.
            raise ValueError("The table argument is required")
        self._add_layer(layer)
        return layer

    def _add_layer(self, layer):
        if layer in self._layers:
            raise ValueError("layer already exists in layer manager")
        self._layers.append(layer)
        layer._manager = self

    def remove_layer(self, layer):
        if layer not in self._layers:
            raise ValueError("layer not in layer manager")
        layer.remove()
        # By this point, the call to remove() above may already have resulted
        # in the layer getting removed, so we check first if it's still present.
        if layer in self._layers:
            self._layers.remove(layer)

    def __len__(self):
        return len(self._layers)

    def __iter__(self):
        for layer in self._layers:
            yield layer

    def __getitem__(self, item):
        return self._layers[item]

    def __str__(self):
        if len(self) == 0:
            return 'Layer manager with no layers'
        else:
            s = 'Layer manager with {0} layers:\n\n'.format(len(self))
            for ilayer, layer in enumerate(self._layers):
                s += '  [{0}]: {1}\n'.format(ilayer, layer)
            return s

    __repr__ = __str__


class TableLayer(HasTraits):
    """
    A layer where the data is stored in an :class:`~astropy.table.Table`
    """

    lon_att = Unicode(help='The column to use for the longitude').tag(wwt='lngColumn')
    lon_unit = Any(help='The units to use for longitude').tag(wwt='raUnits')
    lat_att = Unicode(help='The column to use for the latitude').tag(wwt='latColumn')

    alt_att = Unicode(help='The column to use for the altitude').tag(wwt='altColumn')
    alt_unit = Any(help='The units to use for the altitude').tag(wwt='altUnit')
    alt_type = Unicode(help='The type of altitude').tag(wwt='altType')

    cmap_att = Unicode(help='The column to use for the colormap').tag(wwt='colorMapColumn')

    size_att = Unicode(help='The column to use for the size').tag(wwt='sizeColumn')
    size_scale = Float(10, help='The factor by which to scale the size of the points').tag(wwt='scaleFactor')

    color = Color('white', help='The color of the markers').tag(wwt='color')
    opacity = Float(1, help='The opacity of the markers').tag(wwt='opacity')

    # TODO: support:
    # xAxisColumn
    # yAxisColumn
    # zAxisColumn
    # xAxisReverse
    # yAxisReverse
    # zAxisReverse

    @validate('lon_unit')
    def _check_lon_unit(self, proposal):
        # Pass the proposal to Unit - this allows us to validate the unit,
        # and allows strings to be passed.
        unit = u.Unit(proposal['value'])
        unit = pick_unit_if_available(unit, VALID_LON_UNITS)
        if unit in VALID_LON_UNITS:
            return unit
        else:
            raise ValueError('lon_unit should be one of {0}'.format('/'.join(sorted(str(x) for x in VALID_LON_UNITS))))

    @validate('alt_unit')
    def _check_alt_unit(self, proposal):
        # Pass the proposal to Unit - this allows us to validate the unit,
        # and allows strings to be passed.
        with u.imperial.enable():
            unit = u.Unit(proposal['value'])
        unit = pick_unit_if_available(unit, VALID_ALT_UNITS)
        if unit in VALID_ALT_UNITS:
            return unit
        else:
            raise ValueError('alt_unit should be one of {0}'.format('/'.join(sorted(str(x) for x in VALID_ALT_UNITS))))

    @validate('alt_type')
    def _check_alt_type(self, proposal):
        if proposal['value'] in VALID_ALT_TYPES:
            return proposal['value']
        else:
            raise ValueError('alt_type should be one of {0}'.format('/'.join(str(x) for x in VALID_ALT_TYPES)))

    @observe('alt_att')
    def _on_alt_att_change(self, *value):
        # Check if we can set the unit of the altitude automatically
        if len(self.alt_att) == 0:
            return
        column = self.table[self.alt_att]
        unit = pick_unit_if_available(column.unit, VALID_ALT_UNITS)
        if unit in VALID_ALT_UNITS:
            self.alt_unit = unit
        elif unit is not None:
            warnings.warn('Column {0} has units of {1} but this is not a valid '
                          'unit of altitude - set the unit directly with '
                          'alt_unit'.format(self.alt_att, unit), UserWarning)

    @observe('lon_att')
    def _on_lon_att_change(self, *value):
        # Check if we can set the unit of the altitude automatically
        if len(self.lon_att) == 0:
            return
        column = self.table[self.lon_att]
        unit = pick_unit_if_available(column.unit, VALID_LON_UNITS)
        if unit in VALID_LON_UNITS:
            self.lon_unit = unit
        elif unit is not None:
            warnings.warn('Column {0} has units of {1} but this is not a valid '
                          'unit of longitude - set the unit directly with '
                          'lon_unit'.format(self.lon_att, unit), UserWarning)

    def __init__(self, parent=None, table=None, frame=None, **kwargs):

        # TODO: need to validate reference frame
        self.table = table
        self.frame = frame

        self.parent = parent
        self.id = str(uuid.uuid4())

        # Attribute to keep track of the manager, so that we can notify the
        # manager if a layer is removed.
        self._manager = None
        self._removed = False

        self._initialize_layer()

        # Force defaults
        self._on_trait_change({'name': 'size_scale', 'new': self.size_scale})
        self._on_trait_change({'name': 'color', 'new': self.color})
        self._on_trait_change({'name': 'opacity', 'new': self.opacity})

        self.observe(self._on_trait_change, type='change')

        if any(key not in self.trait_names() for key in kwargs):
            raise KeyError('a key doesn\'t match any layer trait name')

        super(TableLayer, self).__init__(**kwargs)

        lon_guess, lat_guess = guess_lon_lat_columns(self.table.colnames)

        if 'lon_att' not in kwargs:
            self.lon_att = lon_guess or self.table.colnames[0]

        if 'lat_att' not in kwargs:
            self.lat_att = lat_guess or self.table.colnames[1]

    @property
    def _table_b64(self):

        # TODO: We need to make sure that the table has ra/dec columns since
        # WWT absolutely needs that upon creation.

        s = StringIO()
        self.table.write(s, format='ascii.basic', delimiter=',', comment=False)
        s.seek(0)

        # Enforce Windows line endings
        # TODO: check if this needs to be different on Windows
        csv = s.read().replace('\n', '\r\n')

        return b64encode(csv.encode('ascii', errors='replace')).decode('ascii')

    def _initialize_layer(self):
        self.parent._send_msg(event='table_layer_create',
                              id=self.id, table=self._table_b64, frame=self.frame)

    def update_data(self, table=None):
        """
        Update the underlying data.
        """
        self.table = table
        self.parent._send_msg(event='table_layer_update', id=self.id, table=self._table_b64)

        if len(self.alt_att) > 0:
            if self.alt_att in self.table.colnames:
                self._on_alt_att_change()
            else:
                self.alt_att = ''

        lon_guess, lat_guess = guess_lon_lat_columns(self.table.colnames)

        if self.lon_att in self.table.colnames:
            self._on_lon_att_change()
        else:
            self.lon_att = lon_guess or self.table.colnames[0]

        if self.lat_att not in self.table.colnames:
            self.lat_att = lat_guess or self.table.colnames[1]

    def remove(self):
        """
        Remove the layer.
        """
        if self._removed:
            return
        self.parent._send_msg(event='table_layer_remove', id=self.id)
        self._removed = True
        if self._manager is not None:
            self._manager.remove_layer(self)

    def _on_trait_change(self, changed):
        # This method gets called anytime a trait gets changed. Since this class
        # gets inherited by the Jupyter widgets class which adds some traits of
        # its own, we only want to react to changes in traits that have the wwt
        # metadata attribute (which indicates the name of the corresponding WWT
        # setting).
        wwt_name = self.trait_metadata(changed['name'], 'wwt')
        if wwt_name is not None:
            value = changed['new']
            if changed['name'] == 'alt_unit':
                value = VALID_ALT_UNITS[self._check_alt_unit({'value': value})]
            elif changed['name'] == 'lon_unit':
                value = VALID_LON_UNITS[self._check_lon_unit({'value': value})]
            self.parent._send_msg(event='table_layer_set',
                                  id=self.id,
                                  setting=wwt_name,
                                  value=value)

    def __str__(self):
        return 'TableLayer with {0} markers'.format(len(self.table))

    def __repr__(self):
        return '<{0}>'.format(str(self))



from bs4 import BeautifulSoup
import requests

__all__ = ['WWTException', 'get_soup', 'handle_response', 'parse_kwargs']


class WWTException(Exception):
    pass


def get_soup(url, params):
    u = requests.get(url, params=params)
    soup = BeautifulSoup(u.content, "xml")
    return soup, u.content


def handle_response(resp_str):
    soup = BeautifulSoup(resp_str, "xml")
    try:
        success = soup.LayerApi.Status.string
        if success != "Success":
            raise WWTException(success)
    except AttributeError:
        soup2 = BeautifulSoup(resp_str)
        error = soup2.html.body.h2.string
        raise WWTException(error)


def parse_kwargs(params, kwargs):
    if "date_time" in kwargs:
        params["datetime"] = kwargs["date_time"]
    if "time_rate" in kwargs:
        params["timerate"] = str(kwargs["time_rate"])
    if "fly_to" in kwargs:
        params["flyto"] = ",".join([str(i) for i in kwargs["fly_to"]])
    if "instant" in kwargs:
        params["instant"] = str(kwargs["instant"])
    if "auto_loop" in kwargs:
        params["autoloop"] = str(kwargs["auto_loop"])



import requests

from .misc import WWTException, handle_response, parse_kwargs, get_soup

__all__ = ['WWTLayer']


class WWTLayer(object):
    """
    A ``WWTLayer`` object corresponding to a layer in WWT's Layer Manager.

    Parameters
    ----------
    name : `str`
        The name of the layer.
    id : `str`
        The ID of the layer.
    field : `list`
        The fields in the layer as a list of strings.
    wwt : `~pywwt.windows.WWTWindowsClient`
        The WWT client where this layer exists.
    """
    def __init__(self, name, id, fields, wwt):
        self.name = name
        self.id = id
        self.wwt = wwt
        self.exists = True
        self.fields = fields

    def _check_exists(self):
        if not self.exists:
            raise WWTException("This layer has been deleted!")

    def set_property(self, property_name, property_value, **kwargs):
        """
        Set a single property. For a list of properties see:

        https://worldwidetelescope.gitbooks.io/worldwide-telescope-layer-control-api/content/lcapicommands.html#table-of-properties

        Parameters
        ----------
        property_name : `str`
            Name of the property to be set.
        property_value : `str`
            Value the property should take.
        kwargs
            Additional standard keyword arguments
        """
        self._check_exists()
        params = {}
        params["cmd"] = "setprop"
        params["id"] = self.id
        params["propname"] = property_name
        params["propvalue"] = property_value
        parse_kwargs(params, kwargs)
        u = requests.post(self.wwt.wwt_url, params=params)
        prop_str = u.content
        handle_response(prop_str)

    def set_properties(self, props_dict, **kwargs):
        """
        Set the properties of the layer. For a list of properties
        see:

        https://worldwidetelescope.gitbooks.io/worldwide-telescope-layer-control-api/content/lcapicommands.html#table-of-properties

        Parameters
        ----------
        props_dict : `dict`
            Dictionary of property_name: property_value pairs.
        kwargs
            Additional standard keyword arguments
        """
        self._check_exists()
        props_string = "<?xml version='1.0' encoding='UTF-8'?><LayerApi><Layer "
        for key, value in list(props_dict.items()):
            props_string += "%s=\"%s\" " % (key, value)
        props_string += " /></LayerApi>"
        params = {}
        params["cmd"] = "setprops"
        params["id"] = self.id
        parse_kwargs(params, kwargs)
        u = requests.post(self.wwt.wwt_url, params=params, data=props_string)
        props_str = u.content
        handle_response(props_str)

    def get_property(self, property_name):
        """
        Return a property. For a list of properties see:

        https://worldwidetelescope.gitbooks.io/worldwide-telescope-layer-control-api/content/lcapicommands.html#table-of-properties

        Parameters
        ----------
        property_name : `str`
            The name of the property to be returned.
        """
        self._check_exists()
        params = {}
        params["cmd"] = "getprop"
        params["id"] = self.id
        params["propname"] = property_name
        soup, resp = get_soup(self.wwt.wwt_url, params)
        handle_response(resp)
        return soup.Layer.attrs[property_name]

    def get_properties(self):
        """
        Return all the properties of the layer. For a list of properties
        see:

        https://worldwidetelescope.gitbooks.io/worldwide-telescope-layer-control-api/content/lcapicommands.html#table-of-properties

        Returns
        -------
        properties : `dict`
            All of the properties in a dict of key: value pairs.
        """
        self._check_exists()
        params = {}
        params["cmd"] = "getprops"
        params["id"] = self.id
        soup, resp = get_soup(self.wwt.wwt_url, params)
        handle_response(resp)
        return soup.Layer.attrs

    def update(self, data=None, name=None,
               no_purge=False, purge_all=False,
               show=True, **kwargs):
        """
        Update the layer.

        Parameters
        ----------
        data : `dict`, optional
            New data to be added to the layer as a dictionary of Numpy arrays
        name : `str`, optional
            A new name for the layer, if it needs to be renamed
        no_purge : `bool`, optional
            The sending of an update command will delete events that occur
            before the start time of any events in the update and that have
            already decayed. Set this flag to `True` if the event data should
            not be deleted.
        purge_all : `bool`, optional
            Delete all data.
        show : `bool`, optional
            Whether to show or hide the layer.
        kwargs
            Additional standard keyword arguments
        """
        self._check_exists()
        params = {}
        params["cmd"] = "update"
        params["id"] = self.id
        params["hasheader"] = "False"
        params["name"] = name
        params["purgeall"] = str(purge_all).lower()
        params["nopurge"] = str(no_purge).lower()
        params["show"] = str(show).lower()
        parse_kwargs(params, kwargs)
        data_string = ""
        if data is not None:
            nevents = len(data[self.fields[0]])
            for i in range(nevents):
                data_string += "\t".join([str(data[k][i]) for k in self.fields]) + "\n"
        u = requests.post(self.wwt.wwt_url, params=params, data=data_string)
        update_str = u.content
        handle_response(update_str)

    def activate(self, **kwargs):
        """
        Highlight this layer in the layer manager.

        Parameters
        ----------
        kwargs
            Additional standard keyword arguments
        """
        self._check_exists()
        params = {}
        params["cmd"] = "activate"
        params["id"] = self.id
        parse_kwargs(params, kwargs)
        u = requests.get(self.wwt.wwt_url, params=params)
        layer_str = u.content
        handle_response(layer_str)

    def delete(self):
        """
        Delete the layer.
        """
        self._check_exists()
        params = {}
        params["cmd"] = "delete"
        params["id"] = self.id
        u = requests.get(self.wwt.wwt_url, params=params)
        layer_str = u.content
        handle_response(layer_str)
        self.exists = False

    def __repr__(self):
        return "<WWTLayer \"%s\" with ID %s on client at %s>" % (
            self.name, self.id, self.wwt.host)

    def __str__(self):
        return self.name



# This file contains the defintion of the Qt widget. Most of the code here
# deals with differences between WebEngine and WebKit (either of which may be
# available in Qt) and also deals with figuring out how we know once WWT is
# set up.

from __future__ import absolute_import, division, print_function

import os
import json
import time

from qtpy.QtWebEngineWidgets import QWebEngineView, QWebEnginePage, WEBENGINE
from qtpy import QtWidgets, QtGui, QtCore

from .app import get_qapp
from .core import BaseWWTWidget
from .logger import logger
from .data_server import get_data_server

__all__ = ['WWTQtClient']

WWT_JSON_FILE = os.path.join(os.path.dirname(__file__), 'nbextension', 'static', 'wwt_json_api.js')

with open(WWT_JSON_FILE) as f:
    WWT_JSON = f.read()

WWT_HTML_FILE = os.path.join(os.path.dirname(__file__), 'nbextension', 'static', 'wwt.html')

with open(WWT_HTML_FILE) as f:
    WWT_HTML = f.read()


class WWTQWebEnginePage(QWebEnginePage):
    """
    Subclass of QWebEnginePage that can check when WWT is ready for
    commands.
    """

    wwt_ready = QtCore.Signal()

    def __init__(self, parent=None):
        super(WWTQWebEnginePage, self).__init__(parent=parent)
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self._check_ready)
        self._timer.start(500)
        self._check_running = False
        if not WEBENGINE:
            self._frame = self.mainFrame()

    if WEBENGINE:

        def _wwt_ready_callback(self, result):
            if result == 1:
                self._timer.stop()
                self.wwt_ready.emit()
            self._check_running = False

        def javaScriptConsoleMessage(self, level=None, message=None,
                                     line_number=None, source_id=None):
            if not self._check_running or 'wwt_ready' not in message:
                print('{0} (level={1}, line_number={2}, source_id={3})'.format(message, level, line_number, source_id))

        def _check_ready(self):
            if not self._check_running:
                self._check_running = True
                super(WWTQWebEnginePage, self).runJavaScript('wwt_ready;', self._wwt_ready_callback)

        def _process_js_response(self, result):
            self._js_response_received = True
            self._js_response = result

        def runJavaScript(self, code, asynchronous=True):
            app = get_qapp()
            if asynchronous:
                super(WWTQWebEnginePage, self).runJavaScript(code)
            else:
                self._js_response_received = False
                self._js_response = None
                super(WWTQWebEnginePage, self).runJavaScript(code, self._process_js_response)
                while not self._js_response_received:
                    app.processEvents()
                return self._js_response

    else:

        def javaScriptConsoleMessage(self, message=None, line_number=None,
                                     source_id=None):
            if 'wwt_ready' not in message:
                print('{0} (line_number={1}, source_id={2})'.format(message, line_number, source_id))

        def runJavaScript(self, code, asynchronous=False):
            return self._frame.evaluateJavaScript(code)

        def _check_ready(self):
            result = self.runJavaScript('wwt_ready;')
            if result == 1:
                self._timer.stop()
                self.wwt_ready.emit()


class WWTQtWidget(QtWidgets.QWidget):

    def __init__(self, url=None, parent=None):

        super(WWTQtWidget, self).__init__(parent=parent)

        self.web = QWebEngineView()
        self.page = WWTQWebEnginePage()
        self.page.setView(self.web)
        self.web.setPage(self.page)
        self.web.setUrl(QtCore.QUrl(url))

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        layout.addWidget(self.web)

        self._wwt_ready = False
        self._js_queue = ""

        self.page.wwt_ready.connect(self._on_wwt_ready)

    def send_msg(self, **kwargs):
        msg = json.dumps(kwargs)
        return self._run_js("wwt_apply_json_message(wwt, {0})".format(msg))

    def _on_wwt_ready(self):
        self._run_js(WWT_JSON)
        self._wwt_ready = True
        self._run_js(self._js_queue, asynchronous=True)
        self._js_queue = ""

    def _run_js(self, js, asynchronous=True):
        if not js:
            return
        if self._wwt_ready:
            logger.debug('Running javascript: %s' % js)
            return self.page.runJavaScript(js, asynchronous=asynchronous)
        else:
            logger.debug('Caching javascript: %s' % js)
            self._js_queue += js + '\n'


class WWTQtClient(BaseWWTWidget):
    """
    A client to create and drive the Qt widget.

    Parameters
    ----------
    block_until_ready : `bool`
        Tells Python to wait for WorldWide Telescope to open before
        proceeding with any following script (default: `True`).

    size : `tuple`
        Sets size of widget in pixels (default: (600, 600)).
    """

    def __init__(self, block_until_ready=False, size=None):

        app = get_qapp()

        self._data_server = get_data_server()
        self._data_server.serve_file(WWT_JSON_FILE, real_name=True)
        wwt_url = self._data_server.serve_file(WWT_HTML_FILE, real_name=True)

        self.widget = WWTQtWidget(url=wwt_url)
        if size is not None:
            self.widget.resize(*size)
        self.widget.show()

        super(WWTQtClient, self).__init__()

        if block_until_ready:
            while True:
                app.processEvents()
                if self.widget._wwt_ready:
                    break

    def wait(self, duration=None):
        """
        Prevents WorldWide Telescope from closing once Python reaches the
        end of a given script.

        Parameters
        ----------
        duration : int or float or None
            How many seconds to wait for. By default, this waits until the
            Qt window is closed.
        """
        app = get_qapp()
        if duration is None:
            app.exec_()
        else:
            time1 = time.time()
            while time.time() - time1 < duration:
                app.processEvents()

    def _send_msg(self, asynchronous=True, **kwargs):
        msg = json.dumps(kwargs)
        return self.widget._run_js("wwt_apply_json_message(wwt, {0});".format(msg), asynchronous=asynchronous)

    def load_fits_data(self, filename):
        """
        Load a FITS file.

        Parameters
        ----------
        filename : str
            The filename of the FITS file to display.
        """
        self._validate_fits_data(filename)
        url = self._data_server.serve_file(filename, extension='.fits')
        self._send_msg(event='load_fits', url=url)

    def render(self, filename):
        """
        Saves a screenshot of the viewer's current state.

        Parameters
        ----------
        filename : `str`
            The desired name of the image file to be saved.
        """
        image = QtGui.QImage(self.widget.size(), QtGui.QImage.Format_RGB32)
        painter = QtGui.QPainter(image)
        self.widget.render(painter)
        image.save(filename)
        painter.end()



import pytest

from astropy.table import Table
from astropy import units as u

from ..qt import WWTQtClient
from .test_qt_widget import check_silent_output

from ..layers import TableLayer, guess_lon_lat_columns


class TestLayers:

    def setup_class(self):
        self.widget = WWTQtClient(block_until_ready=True)
        self.table = Table()
        self.table['flux'] = [2, 3, 4]
        self.table['dec'] = [4, 5, 6]
        self.table['ra'] = [1, 2, 3] * u.deg

    def test_add_and_remove_layer(self, capsys):

        assert len(self.widget.layers) == 0
        assert str(self.widget.layers) == 'Layer manager with no layers'
        layer1 = self.widget.layers.add_data_layer(table=self.table)
        assert len(self.widget.layers) == 1
        assert str(self.widget.layers) == ('Layer manager with 1 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n')

        layer2 = self.widget.layers.add_data_layer(table=self.table)

        assert len(self.widget.layers) == 2
        assert str(self.widget.layers) == ('Layer manager with 2 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n'
                                           '  [1]: TableLayer with 3 markers\n')

        assert self.widget.layers[0] is layer1
        assert self.widget.layers[1] is layer2

        # Test iteration
        for layer in self.widget.layers:
            assert isinstance(layer, TableLayer)

        layer1.remove()

        assert len(self.widget.layers) == 1
        assert str(self.widget.layers) == ('Layer manager with 1 layers:\n\n'
                                           '  [0]: TableLayer with 3 markers\n')

        self.widget.layers.remove_layer(layer2)

        assert len(self.widget.layers) == 0
        assert str(self.widget.layers) == 'Layer manager with no layers'

        self.widget.wait(1)
        # check_silent_output(capsys)

    def test_alt_unit(self):

        layer = self.widget.layers.add_data_layer(table=self.table)

        # Using a string
        layer.alt_unit = 'm'

        # Using a string of an imperial unit
        layer.alt_unit = 'inch'

        # Using an astropy unit
        layer.alt_unit = u.km

        # Using a unit that is equal but not identical to one of the accepted ones
        layer.alt_unit = u.def_unit('same_as_km', 1000 * u.m)

        # Using an invalid string
        with pytest.raises(ValueError) as exc:
            layer.alt_unit = 'banana'
        assert exc.value.args[0].strip() == "'banana' did not parse as unit: At col 0, banana is not a valid unit."

        # Using an unsupported unit
        with pytest.raises(ValueError) as exc:
            layer.alt_unit = u.kg
        assert exc.value.args[0].strip() == "alt_unit should be one of AU/Mpc/ft/inch/km/lyr/m/mi/pc"

        # Using a non-equal custom unit
        with pytest.raises(ValueError) as exc:
            layer.alt_unit = u.def_unit('same_as_half_km', 500 * u.m)
        assert exc.value.args[0].strip() == "alt_unit should be one of AU/Mpc/ft/inch/km/lyr/m/mi/pc"

    def test_lon_unit(self):

        layer = self.widget.layers.add_data_layer(table=self.table)

        # Using a string
        layer.lon_unit = 'deg'

        # Using an astropy unit
        layer.lon_unit = u.hourangle

        # Using a unit that is equal but not identical to one of the accepted ones
        layer.lon_unit = u.def_unit('same_as_deg', 3600 * u.arcsec)

        # Using an invalid string
        with pytest.raises(ValueError) as exc:
            layer.lon_unit = 'banana'
        assert exc.value.args[0].strip() == "'banana' did not parse as unit: At col 0, banana is not a valid unit."

        # Using an unsupported unit
        with pytest.raises(ValueError) as exc:
            layer.lon_unit = u.kg
        assert exc.value.args[0].strip() == "lon_unit should be one of deg/h/hourangle"

        # Using a non-equal custom unit
        with pytest.raises(ValueError) as exc:
            layer.lon_unit = u.def_unit('same_as_arcmin', 60 * u.arcsec)
        assert exc.value.args[0].strip() == "lon_unit should be one of deg/h/hourangle"

    def test_alt_type(self):

        layer = self.widget.layers.add_data_layer(table=self.table)

        layer.alt_type = 'depth'

        with pytest.raises(ValueError) as exc:
            layer.alt_type = 'time'
        assert exc.value.args[0].strip() == "alt_type should be one of depth/altitude/distance/seaLevel/terrain"

    def test_auto_alt_unit(self):

        self.table['altitude'] = [1, 4, 5] * u.au
        self.table['altitude2'] = [1, 4, 5] * u.def_unit('same_as_km', 1000 * u.m)
        self.table['flux'].unit = u.kg

        layer = self.widget.layers.add_data_layer(table=self.table)

        assert layer.alt_att == ''
        assert layer.alt_unit is None

        layer.alt_att = 'altitude'
        assert layer.alt_unit is u.au

        layer.alt_att = 'altitude2'
        assert layer.alt_unit is u.km

        expected_warning = ('Column flux has units of kg but this is not a '
                            'valid unit of altitude - set the unit directly with '
                            'alt_unit')

        with pytest.warns(UserWarning, match=expected_warning):
            layer.alt_att = 'flux'

    def test_auto_lon_unit(self):

        self.table['longitude'] = [1, 4, 5] * u.hour
        self.table['longitude2'] = [1, 4, 5] * u.def_unit('same_as_deg', 3600 * u.arcsec)
        self.table['flux'].unit = u.kg

        layer = self.widget.layers.add_data_layer(table=self.table)

        assert layer.lon_att == 'ra'
        assert layer.lon_unit is u.deg

        layer.lon_att = 'longitude'
        assert layer.lon_unit is u.hour

        layer.lon_att = 'longitude2'
        assert layer.lon_unit is u.deg

        expected_warning = ('Column flux has units of kg but this is not a '
                            'valid unit of longitude - set the unit directly with '
                            'lon_unit')

        with pytest.warns(UserWarning, match=expected_warning):
            layer.lon_att = 'flux'

    def test_update_data(self):

        self.table['flux'].unit = 'm'
        layer = self.widget.layers.add_data_layer(table=self.table,
                                                  lon_att='ra', lat_att='dec', alt_att='flux')

        assert layer.lon_att == 'ra'
        assert layer.lon_unit is u.deg
        assert layer.lat_att == 'dec'
        assert layer.alt_att == 'flux'
        assert layer.alt_unit is u.m

        # Replace with a table with the same column names but different units
        # for the lon and alt
        table = Table()
        table['ra'] = [1, 2, 3] * u.hourangle
        table['dec'] = [4, 5, 6]
        table['flux'] = [2, 3, 4] * u.km
        layer.update_data(table=table)

        assert layer.lon_att == 'ra'
        assert layer.lon_unit is u.hourangle
        assert layer.lat_att == 'dec'
        assert layer.alt_att == 'flux'
        assert layer.alt_unit is u.km

        # Replace with a table with different column names
        table = Table()
        table['a'] = [1, 2, 3] * u.deg
        table['b'] = [4, 5, 6]
        table['c'] = [2, 3, 4] * u.au
        layer.update_data(table=table)

        assert layer.lon_att == 'a'
        assert layer.lon_unit is u.deg
        assert layer.lat_att == 'b'
        assert layer.alt_att == ''


CASES = [[('flux', 'dec', 'ra'), ('ra', 'dec')],
         [('mass', 'lat', 'lon'), ('lon', 'lat')],
         [('a', 'lng', 'b', 'lat'), ('lng', 'lat')],
         [('flux', 'ra', 'radius', 'dec'), ('ra', 'dec')],
         [('FLUX', 'DECJ2000', 'RAJ2000'), ('RAJ2000', 'DECJ2000')],
         [('DISTANCE', 'LON1', 'LAT1'), ('LON1', 'LAT1')],
         [('flux', 'lng2', 'lat2', 'lng1', 'lat1'), (None, None)],
         [('ra', 'ra', 'dec'), (None, None)]]


@pytest.mark.parametrize(('colnames', 'expected'), CASES)
def test_guess_lon_lat_columns(colnames, expected):
    assert guess_lon_lat_columns(colnames) == expected



import os
import time
import socket
import logging
from hashlib import md5
from threading import Thread

__all__ = ['get_data_server']


def get_data_server(verbose=True):
    """
    This starts up a flask server and returns a handle to a DataServer
    object which can be used to register files to serve.
    """

    from flask import Flask
    from flask_cors import CORS

    class FlaskWrapper(Flask):

        port = None
        host = None

        def run(self, *args, **kwargs):
            self.host = kwargs.get('host', None)
            self.port = kwargs.get('port', None)
            try:
                super(FlaskWrapper, self).run(*args, **kwargs)
            finally:
                self.host = None
                self.port = None

    app = FlaskWrapper('DataServer')
    CORS(app)
    if verbose:
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    class DataServer(object):

        def __init__(self):
            self._files = {}
            self._thread = Thread(target=self.start_app)
            self._thread.daemon = True
            self._thread.start()
            self._app = app
            while self._app.host is None and self._app.port is None:
                time.sleep(0.1)

        @property
        def port(self):
            return self._app.port

        @property
        def host(self):
            return self._app.host

        def start_app(self):
            host = socket.gethostbyname('localhost')
            for port in range(8000, 9000):
                try:
                    return app.run(host=host, port=port)
                except Exception:
                    pass
            raise Exception("Could not start up data server")

        def serve_file(self, filename, real_name=True, extension=''):
            with open(filename, 'rb') as f:
                content = f.read()
            if real_name:
                hash = os.path.basename(filename)
            else:
                hash = md5(content).hexdigest() + extension
            self._files[hash] = os.path.abspath(filename)
            return 'http://' + self.host + ':' + str(self.port) + '/data/' + hash

        def get_file_contents(self, hash):
            with open(self._files[hash], 'rb') as f:
                return f.read()

    ds = DataServer()

    @app.route("/data/<hash>")
    def data(hash):
        return ds.get_file_contents(hash)

    return ds



from astropy import units as u
from traitlets import HasTraits, validate

from .traits import Bool, Int

__all__ = ['SolarSystem']


class SolarSystem(HasTraits):
    """
    A supplemental class that enables tab-completion for settings
    associated with solar system mode.
    """

    def __init__(self, base_wwt_widget):
        super(SolarSystem, self).__init__()
        self.base_widget = base_wwt_widget
        self.observe(self._on_trait_change, type='change')

    def _on_trait_change(self, changed):
        # This method gets called anytime a trait gets changed. Since this class
        # gets inherited by the Jupyter widgets class which adds some traits of
        # its own, we only want to react to changes in traits that have the wwt
        # metadata attribute (which indicates the name of the corresponding WWT
        # setting).
        wwt_name = self.trait_metadata(changed['name'], 'wwt')
        new_value = changed['new']
        if wwt_name is not None:
            if isinstance(new_value, u.Quantity):
                new_value = new_value.value

            self.base_widget._send_msg(event='setting_set',
                                setting=wwt_name,
                                value=new_value)

    #cmb = Bool(False, help='Whether to show the cosmic microwave background in solar system mode (`bool`)').tag(wwt='solarSystemCMB') ###
    #cosmos = Bool(False, help='Whether to show data from the SDSS Cosmos data set (`bool`)').tag(wwt='solarSystemCosmos') ###
    #display = Bool(False, help='Whether to show the solar system while in solar system mode (`bool`)').tag(wwt='solarSystemOverlays') ###
    lighting = Bool(True,
                    help='Whether to show the lighting effect of the Sun on the'
                         ' solar system (`bool`)').tag(wwt='solarSystemLighting')
    milky_way = Bool(True, help='Whether to show the galactic bulge in the '
                                'background in solar system mode '
                                '(`bool`)').tag(wwt='solarSystemMilkyWay')
    #multi_res = Bool(False, help='Whether to show the multi-resolution textures for planets where available (`bool`)').tag(wwt='solarSystemMultiRes') ###
    minor_orbits = Bool(False, help='Whether to show the orbits of minor planets in solar system mode (`bool`)').tag(wwt='solarSystemMinorOrbits')
    #minor_planets = Bool(False, help='Whether to show minor planets in solar system mode (`bool`)').tag(wwt='solarSystemMinorPlanets') ###
    orbits = Bool(True,
                  help='Whether to show orbit paths when the solar system is '
                       'displayed (`bool`)').tag(wwt='solarSystemOrbits')
    objects = Bool(True,
                   help='Whether to show the objects of the solar system in '
                        'solar system mode (`bool`)').tag(wwt='solarSystemPlanets')
    scale = Int(1, help='Specifies how to scale objects\' size in solar '
                        'system mode, with 1 as actual size and 100 as the '
                        'maximum (`int`)').tag(wwt='solarSystemScale')
    stars = Bool(True, help='Whether to show background stars in solar system '
                            'mode (`bool`)').tag(wwt='solarSystemStars')

    @validate('scale')
    def _validate_scale(self, proposal):
        if 1 <= proposal['value'] <= 100:
            return str(proposal['value'])
        else:
            raise ValueError('scale takes integers from 1-100')

    def track_object(self, obj):
        """
        Focus the viewer on a particular object while in solar system mode.
        Available objects include the Sun, the planets, the Moon, Jupiter's
        Galilean moons, and Pluto.

        Parameters
        ----------
        obj : `str`
            The desired solar system object.
        """
        obj = obj.lower()
        mappings = {'sun': 0, 'mercury': 1, 'venus': 2, 'mars': 3, 'jupiter': 4,
                    'saturn': 5, 'uranus': 6, 'neptune': 7, 'pluto': 8,
                    'moon': 9, 'io': 10, 'europa': 11, 'ganymede': 12,
                    'callisto': 13, 'ioshadow': 14, 'europashadow': 15,
                    'ganymedeshadow': 16, 'callistoshadow': 17,
                    'suneclipsed': 18, 'earth': 19}

        if obj in mappings:
            self.base_widget._send_msg(event='track_object', code=mappings[obj])
        else:
            raise ValueError('the given object cannot be tracked')



import re

from io import BytesIO
from collections import OrderedDict
from xml.etree.ElementTree import ElementTree

import requests


def get_imagery_layers(url):
    """
    Get the list of available image layers that can be used as background
    or foreground based on the URL to a WTML (WorldWide Telescope image
    collection file).

    Parameters
    ----------
    url : `str`
        The URL of the image collection.
    """

    available_layers = OrderedDict()

    # Get the XML describing the available surveys
    response = requests.get(url)
    assert response.ok
    b = BytesIO(response.content)
    e = ElementTree()
    t = e.parse(b)

    # For now only look at the ImageSets at the root of the
    # XML since these seem to be the main surveys.
    for survey in t.findall('ImageSet'):
        name = survey.attrib['Name']
        thumbnail_url = survey.find('ThumbnailUrl').text
        if not thumbnail_url:
            thumbnail_url = None
        available_layers[name] = {'thumbnail': thumbnail_url}

    return available_layers


class ImageryLayers():
    """
    A supplemental class that standardizes layer names, sorts them by
    bandpass, and makes it possible to tab complete them.
    """

    def __init__(self, layer_list):
        self._layers = {}
        self._spectrum = ['gamma', 'x', 'uv', 'visible',
                          'ir', 'micro', 'radio', 'other']
        self.integers = ['zero', 'one', 'two', 'three', 'four',
                         'five', 'six', 'seven', 'eight', 'nine']
        for band in self._spectrum:
            self._layers[band] = {}

        self._list2dict(layer_list)

    def _list2dict(self, og_list):
        # Helps turn the list of layer names used to initialize the class
        # (og_list) into a dict.
        for layer in og_list:

            if re.search(r'(?i)gamma',layer) is not None:
                self._add2dict(self._layers, layer, 'gamma')
                continue # automatically advance to next iteration

            if re.search(r'(?i)x(-|\s)?ray',layer) is not None:
                self._add2dict(self._layers, layer, 'x')
                continue

            if (re.search(r'(?i)ultra(-|\s)?violet',layer) is not None or
                re.search(r'(?i)[^\d\w]+uv|uv[^\d\w]+',layer) is not None):
                self._add2dict(self._layers, layer, 'uv')
                continue

            if (re.search(r'(?i)optical',layer) is not None or
                re.search(r'(?i)visible',layer) is not None):
                self._add2dict(self._layers, layer, 'visible')
                continue

            if (re.search(r'(?i)infrared',layer) is not None or
                re.search(r'(?i)[^\d\w]+ir|ir[^\d\w]+',layer) is not None):
                self._add2dict(self._layers, layer, 'ir')
                continue

            if (re.search(r'(?i)microwave',layer) is not None or
                re.search(r'(?i)[^\d\w]+cmb|cmb[^\d\w]+',layer) is not None):
                self._add2dict(self._layers, layer, 'micro')
                continue

            if re.search(r'(?i)radio',layer) is not None:
                self._add2dict(self._layers, layer, 'radio')
                continue

            self._add2dict(self._layers, layer, 'other')

    def _add2dict(self, diction, full_layer, bandpass):
        # Handles a layer's (full_layer) actual addition to the master
        # dict of layers (diction) according to its bandpass.
        suffix = ''
        short = self._shorten(full_layer) + suffix

        while short in diction[bandpass]:
            if suffix:
                suffix += 1
            else:
                suffix = 1

            short += str(suffix)

        diction[bandpass][short] = {}
        diction[bandpass][short]['full_name'] = full_layer

    def _shorten(self, string):
        # Unlocks tab completion by shortening a full layer's name
        # (string) to a valid Python name based on its first word.
        cut_left = re.search(r'^[_\W]+', string)
        if cut_left is not None:
            string = string[cut_left.end():]

        cut_right = re.search(r'[_\W]', string)
        if cut_right is not None:
            string = string[:cut_right.start()].lower()

        digit = re.search(r'^\d', string)
        if digit is not None:
            for i, num in enumerate(self.integers, 0):
                if str(i) == digit.group(0):
                    string = string[:digit.start()] + num + string[digit.end():]
                    break

        return string

    def __dir__(self):
        return sorted(self._layers.keys())

    def __getattr__(self, band):
        return Bandpass(self._layers[band])


class Bandpass():
    """
    Allows the __getattr__() method from ImageryLayers to reach into the
    inner dict corresponding to layers of a particular bandpass.
    """

    def __init__(self, band):
        self._band = band

    def __dir__(self):
        return sorted(self._band.keys())

    def __getattr__(self, name):
        return self._band[name]['full_name']
